#include <stdio.h>
#include <string>

#include <graphviz/gvc.h>

#include <codegen/AutomataOutput.h>
#include <Syntax2Automata/SyntaxHelpers.h>

#include <common/comm.h>

struct GraphVizContext {
    Agraph_t* pGraph;
    GVC_t* pGvc;
    const common::Automaton* pAutomaton;
};

extern "C" void hack_clearKey();

namespace {

    void buildGraphLegend(GraphVizContext& context)
    {
        Agraph_t* subgraph = agsubg(context.pGraph,"cluster_legend");
        agsafeset(subgraph, "label", "Legend: ", "Legend: ");    
        agset(subgraph, "size", "0.5");
        
        Agnode_t* node = agnode(subgraph, "Initial/final\nstate");
        agsafeset(node, "shape", "circle", "circle");
        agsafeset(node, "fillcolor", "orange", "orange");
        agsafeset(node, "margin", "0.02", "0.02");
        
        
        Agnode_t* node2 = agnode(subgraph, "Operator state");
        agsafeset(node2, "shape", "box", "box");
        agsafeset(node2, "fillcolor", "yellow", "yellow");
        
        Agedge_t* edge = agedge(subgraph, node, node2);
        agsafeset(edge, "color", "white", "white");
        

        node = agnode(subgraph, "Conditional state");
        agsafeset(node, "shape", "parallelogram", "parallelogram");
        agsafeset(node, "fillcolor", "yellow", "yellow");
        
        edge = agedge(subgraph, node2, node);
        agsafeset(edge, "color", "white", "white");

        node2 = agnode(subgraph, "Internal state");
        agsafeset(node2, "shape", "box", "box");
        agsafeset(node2, "fillcolor", "lightgrey", "lightgrey");

        edge = agedge(subgraph, node, node2);
        agsafeset(edge, "color", "white", "white");
        


    }

    
    bool isInternalAction(const std::string& action)
    {
        return (action.find("stack_->") != std::string::npos);
    }

    bool isInternaleState(const common::AutomatonState* state)
    {
        if (state->stateData) {
            return state->stateData->stateType == common::AutomatonStateData::STATE_TYPE_INTERNAL; 
        }
        int nonInternalActions = 0;
        for (unsigned int i = 0; i < state->actions.size(); i++) {
            if (!isInternalAction(state->actions[i])) {
                nonInternalActions++;
            }
        }
        return (nonInternalActions == 0 && state->transitions.size() == 1); // no actions, single transition - this is some kind of internal state
    }

    void setUpStateAttributes(const common::AutomatonState* state, Agnode_t* node)
    {
        agsafeset(node, "margin", "0.3", "0.3");
        agsafeset(node, "fontsize", "14", "14");
        if (isInternaleState(state)) {
            agsafeset(node, "shape", "box", "box");
            agsafeset(node, "fillcolor", "lightgrey", "lightgrey");
            return;
        }
        if (!state->stateData) {
            agsafeset(node, "shape", "box", "box");
            agsafeset(node, "fillcolor", "yellow", "yellow");
            return;
        }

        if (state->stateData->stateType == common::AutomatonStateData::STATE_TYPE_CONDITIONAL) {
            agsafeset(node, "shape", "parallelogram", "parallelogram");
            agsafeset(node, "fillcolor", "yellow", "yellow");
            return;
        }
        if (state->stateData->stateType == common::AutomatonStateData::STATE_TYPE_START_END) {
            agsafeset(node, "shape", "circle", "circle");
            agsafeset(node, "fillcolor", "orange", "orange");
            return;
        }
        agsafeset(node, "shape", "box", "box");
        agsafeset(node, "fillcolor", "yellow", "yellow");
    }

    void init(GraphVizContext& context)
    {
        context.pGvc = gvContext();
    
    }

    void term(GraphVizContext& context)
    {
        gvFreeLayout(context.pGvc, context.pGraph);
        agclose(context.pGraph);
        hack_clearKey();
    }

   
    void process(GraphVizContext& context)
    {
        Agnode_t** node = new Agnode_t*[context.pAutomaton->endState];
        char numBuf[128];
        std::string buff;
        for (int i = 0; i < context.pAutomaton->endState; i++) {                            //creating vertixes in graph
            buff.clear();
            sprintf(numBuf, "State %d\n",  context.pAutomaton->states[i]->stateNumber);
            node[i] = agnode(context.pGraph, numBuf);
            buff.append(numBuf);
            if (context.pAutomaton->states[i]->stateData && context.pAutomaton->states[i]->stateData->stateType == common::AutomatonStateData::STATE_TYPE_AUTOMATON_CALL) {
                buff.append(context.pAutomaton->states[i]->stateData->stringData).append("\n");
            } else {
                for (unsigned int j = 0; j < context.pAutomaton->states[i]->actions.size(); j++) {
                    
                    if (!isInternalAction(context.pAutomaton->states[i]->actions[j])) {
                        buff.append(context.pAutomaton->states[i]->actions[j]).append(" \n");
                    }                
                }
            }
            s2a::cleanString(buff);
            agsafeset(node[i], "label", const_cast<char*>(buff.c_str()), "");
           
            setUpStateAttributes(context.pAutomaton->states[i], node[i])          ;
            
            agsafeset(node[i], "style", "filled", "filled");
            agsafeset(node[i], "fontname", "courier-new", "courier-new");
        }
        Agedge_t* edge;
        
        std::string condition;
        for (int i = 0; i < context.pAutomaton->endState; i++) {                            //creating transitions
            for (unsigned int j = 0; j < context.pAutomaton->states[i]->transitions.size(); j++) {
                edge = agedge(context.pGraph, node[i],  node[context.pAutomaton->states[i]->transitions[j]->to]);
                agsafeset(edge, "color", "black", "black");
                if (context.pAutomaton->states[i]->transitions[j]->isConditional) {
                    agsafeset(edge, "fontname", "courier-new", "courier-new");
                    
                    condition = context.pAutomaton->states[i]->transitions[j]->condition;
                    s2a::cleanString(condition);
                    agsafeset(edge, "label", const_cast<char*>(condition.c_str()), "");
                }
            }    
        }
        delete[] node;
    }

} //anonymous namespace


int output::generateAutomatonScheme(GraphVizContext* context, const common::Automaton* automaton, char* fileName)
{
    
    context->pAutomaton = automaton;
    context->pGraph = agopen(const_cast<char*>(context->pAutomaton->name.c_str()), AGDIGRAPHSTRICT);
    ::process(*context);
    buildGraphLegend(*context);
    gvLayout(context->pGvc, context->pGraph, "dot");
    gvRenderFilename(context->pGvc, context->pGraph, "png", fileName);
    ::term(*context);


    return 0;

}

int output::generateAutomataSystemSchemes(const common::AutomataSystem* system)
{
    char* name;
    GraphVizContext context;
    ::init(context);
    for (unsigned int i = 0; i < system->automata.size(); i++) {
        name = new char[system->automata[i]->name.length() + 32];
        sprintf(name,"Schemes\\%s_scheme.png", system->automata[i]->name.c_str());
        output::generateAutomatonScheme(&context, system->automata[i], name);
        delete[] name;
    }
    return 0;
}