#include <string>

#include <cppunit/extensions/HelperMacros.h>

#include <Syntax2Automata/Syntax2Automata.h>
#include <Syntax2Automata/SyntaxHelpers.h>

#include <common/SyntaxTree.h>
#include <common/Automata.h>
#include <Parser/Parser.h>

class Syntax2AutomataTest : public CPPUNIT_NS::TestFixture
{
	CPPUNIT_TEST_SUITE( Syntax2AutomataTest );
	CPPUNIT_TEST(test_removeDataPrefix);
	CPPUNIT_TEST(test_convertToGlobalName);
	CPPUNIT_TEST(test_singleOperator);
	CPPUNIT_TEST(test_singleVariableDeclaration);
	CPPUNIT_TEST(test_variableDeclarationWithInitialization);
	CPPUNIT_TEST(test_ifSimple);
	CPPUNIT_TEST(test_whileSimple);
	CPPUNIT_TEST_SUITE_END();


	common::SyntaxTreeNode* _procedure;
	common::SyntaxTreeNode* _operator;
	common::SyntaxTreeNode* _semicolon;
	common::SyntaxTreeNode* _definition;
	common::SyntaxTreeNode* _definitionWithInit;
	common::SyntaxTreeNode* _if;

public:
	Syntax2AutomataTest()
	{
		_procedure = new common::SyntaxTreeNode (common::SyntaxTreeNode::NTYPE_PROCEDURE, "main");
		_operator = new common::SyntaxTreeNode(common::SyntaxTreeNode::NTYPE_OPERATOR, "x = 5");
		_semicolon = new common::SyntaxTreeNode (common::SyntaxTreeNode::NTYPE_SEMICOLON, "");
		_definition = new common::SyntaxTreeNode (common::SyntaxTreeNode::NTYPE_VARIABLE_DECLARATION, "int x");
		_definitionWithInit = new common::SyntaxTreeNode (common::SyntaxTreeNode::NTYPE_VARIABLE_DECLARATION, "int x = 5");
		_if = new common::SyntaxTreeNode(common::SyntaxTreeNode::NTYPE_IF, "x == 3");
	}

	~Syntax2AutomataTest()
	{
		delete _procedure;
		delete _operator;
		delete _semicolon;
		delete _definition;
		delete _definitionWithInit;
		delete _if;
	}

	void tearDown()
	{
		_procedure->_left = NULL;
		_procedure->_right = NULL;

		_operator->_left = NULL;
		_operator->_right = NULL;

		_semicolon->_left = NULL;
		_semicolon->_right = NULL;

		_definition->_left = NULL;
		_definition->_right = NULL;

		_definitionWithInit->_left = NULL;
		_definitionWithInit->_right = NULL;

		_if->_left = NULL;
		_if->_right = NULL;
	}

	void test_removeDataPrefix()
	{
		std::string str("data->x");
		s2a::removeDataPrefix(str);
		CPPUNIT_ASSERT(str == "x");
	}

	void test_convertToGlobalName()
	{
		CPPUNIT_ASSERT(s2a::converToGlobalName(std::string("x")) == "x");
		CPPUNIT_ASSERT(s2a::converToGlobalName(std::string("data->x")) == "globalData->x");
	}

	void test_singleOperator()
	{
		_procedure->_left = _operator;
		
		s2a::AutomatonConverter conv;
		conv.process(_procedure, false);
		common::Automaton* a = conv.getAutomaton(_procedure);

		CPPUNIT_ASSERT_EQUAL(3, a->endState);
		CPPUNIT_ASSERT(a->name == "main");
		CPPUNIT_ASSERT_EQUAL(a->states[1]->stateNumber, 1);
		CPPUNIT_ASSERT(a->states[1]->actions.size() == 1);
		CPPUNIT_ASSERT(a->states[1]->transitions.size() == 1);
		CPPUNIT_ASSERT_EQUAL(a->states[1]->actions[0], _operator->_text);

		CPPUNIT_ASSERT(a->reverseStates[1]->actions.size() == 0);
		CPPUNIT_ASSERT(a->reverseStates[1]->transitions.size() == 1);

			
	}

	void test_singleVariableDeclaration()
	{
		_procedure->_left = _definition;
		s2a::AutomatonConverter conv;
		conv.process(_procedure, false);
		common::Automaton* a = conv.getAutomaton(_procedure);

		CPPUNIT_ASSERT(a->data.name == "main");
		CPPUNIT_ASSERT(a->data.variables.size() == 1);
		
		CPPUNIT_ASSERT(a->data.variables[0].getName() == "x");
		CPPUNIT_ASSERT(a->data.variables[0].getType() == "int");
	}

	void test_variableDeclarationWithInitialization()
	{
		_procedure->_left = _definitionWithInit;
		s2a::AutomatonConverter conv;
		conv.process(_procedure, false);
		common::Automaton* a = conv.getAutomaton(_procedure);

		CPPUNIT_ASSERT(a->data.name == "main");
		CPPUNIT_ASSERT(a->data.variables.size() == 1);

		CPPUNIT_ASSERT(a->data.variables[0].getName() == "x");
		CPPUNIT_ASSERT(a->data.variables[0].getType() == "int");

		CPPUNIT_ASSERT_EQUAL(a->endState, 3);
		CPPUNIT_ASSERT(a->name == "main");
		CPPUNIT_ASSERT_EQUAL(1, a->states[1]->stateNumber);
		CPPUNIT_ASSERT(a->states[1]->actions.size() == 2);
		CPPUNIT_ASSERT(a->states[1]->transitions.size() == 1);
		CPPUNIT_ASSERT_EQUAL(_operator->_text, a->states[1]->actions[1]);

		
		CPPUNIT_ASSERT(a->reverseStates[1]->transitions.size() == 1);
	}

	void test_ifSimple()
	{
		_procedure->_left = _if;
		_if->_left = _operator;
		s2a::AutomatonConverter conv;
		conv.process(_procedure, false);
		common::Automaton* a = conv.getAutomaton(_procedure);
		
		CPPUNIT_ASSERT_EQUAL(5, a->endState);
		CPPUNIT_ASSERT_EQUAL(a->states[2]->actions[0], _operator->_text);
		
	}

	void test_whileSimple()
	{
		_procedure->_left = _if;
		_if->_left = _operator;
		s2a::AutomatonConverter conv;
		conv.process(_procedure, false);
		common::Automaton* a = conv.getAutomaton(_procedure);

		CPPUNIT_ASSERT_EQUAL(5, a->endState);
		CPPUNIT_ASSERT_EQUAL(a->states[2]->actions[0], _operator->_text);

	}


};